#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/glut.h>

#include "rigidXform.h"

#include "randj.h"

#define MIN(x,y)	(((x)<(y))?(x):(y))
#define MAX(x,y)	(((x)>(y))?(x):(y))

#define DEG_TO_RAD   (M_PI/180.0)

#define FLOAT_GL GLdouble
#define GLVERTEX glVertex3d

#define MAX_NUM_RENDERED_PIXELS (1024*1280)

static int Debug = 0;

static GLsizei CurrentWindowSize[2] = {907, 907};

#if 0
static GLsizei CurrentViewportSize[2];
#endif


/*

Constructing the case:

	- make a list of line segments with end points randomly distributed 
		through the unit sphere
	- make a list of camera orientations randomly distributed through
		the unit 3-sphere
	- keep camera location at origin
	- get sequence of fovs  :  two degree increments from 2 to 178?
	- don't need to vary near/far
	- construct set of window sizes: Maybe powers of 2? 128, 256, 512, 1024
		then the Rave screen size as a special case?


*/


// Note that this currently assumes a symmetric frustum
typedef struct camera_struct
	{
	double location[3];
	double orientation[4];
	GLdouble fov;
	GLdouble near, far;
	} Camera;

typedef struct view_struct
	{
	int idNum;
	Camera camera;
	GLsizei windowSize[2];
	} View;

typedef double Segment[2][3];


static void
printCamera (FILE *fp, Camera *camera)
	{
	fprintf (fp, "Camera:\n");
	fprintf (fp, "    location %f %f %f\n", 
			camera->location[0],
			camera->location[1],
			camera->location[2]);
	fprintf (fp, "    orientation %f %f %f %f\n",
			camera->orientation[0],
			camera->orientation[1],
			camera->orientation[2],
			camera->orientation[3]);
	fprintf (fp, "    fov %f\n", camera->fov);
	fprintf (fp, "    near %f    far %f\n", camera->near, camera->far);

	}

static void
printView (FILE *fp, View *view)
	{
	fprintf (fp, "VIEW    idNum = %d\n", view->idNum);
	printCamera (fp, &(view->camera));
	fprintf (fp, "windowSize = %d %d\n", 
			view->windowSize[0], view->windowSize[1]);
	fprintf (fp, "\n");
	}

static void 
printSeg (Segment seg)
	{
	printf ("      ( %f %f %f )    ( %f %f %f )\n", 
		seg[0][0],
		seg[0][1],
		seg[0][2],
		seg[1][0],
		seg[1][1],
		seg[1][2]);
	}  // end of printSeg

////////////////////////////////////////////////////////////////

//
// In this section we are going to initialize a bunch of test
// cases for checking the point rendering errors.  This will
// involve generating a set of test points, and test viewing
// parameters.
//

static int NumTestViews = 8;
static View *TestViews = NULL;
//static int NumTestSegments = 100000;
static int NumTestSegments = 100;
// static int NumTestSegments = 1500;
static Segment *TestSegments = NULL;

static void
initRenderingTestCases ()
	{
	int i, j; 
#if 1
	int iv, iwin, ifov, ior;
	int winSize[2][2] = { {128, 128}, {907, 907} };
	double fov[2] = {30.0, 150.0};
	double orientation[2][4];
	double len;
#endif

	if ((TestViews = (View *) malloc(NumTestViews * sizeof (View))) == NULL)
		{
		fprintf (stderr, "Unable to allocate memory for test views.\n");
		exit (-1);
		}

	if ((TestSegments = (Segment *) 
			  malloc (NumTestSegments * sizeof (Segment))) == NULL)
		{
		fprintf(stderr, 
			"Unable to allocate memory for test segments.\n");
		exit (-1);
		}


	// First I'm going to hard code one view


	TestViews[0].windowSize[0] = 200;
	TestViews[0].windowSize[1] = 200;
	TestViews[0].camera.location[0] = 0;
	TestViews[0].camera.location[1] = 0;
	TestViews[0].camera.location[2] = 0;
	TestViews[0].camera.orientation[0] = 0;
	TestViews[0].camera.orientation[1] = 0;
	TestViews[0].camera.orientation[2] = 0;
	TestViews[0].camera.orientation[3] = 1;
	TestViews[0].camera.fov = 90.0;
	// TestViews[0].camera.fov = 168.57881;  // y/z < .1
	TestViews[0].camera.near = 0.001;
	TestViews[0].camera.far = 20.0;


#if 1

	// generate views
	ustart (4339);  // seed for views

	for (ior = 0; ior < 2; ior++)
		{
		len = 0;
		for (j = 0; j < 4; j++)
			{
			orientation[ior][j] = uni ();
			len += orientation[ior][j] * orientation[ior][j];
			}

		len = sqrt (len);

		for (j = 0; j < 4; j++)
			{
			orientation[ior][j] /= len;
			}
		}


	iv = 0;
	for (iwin = 0; iwin < 2; iwin++)
	    {
	    for (ifov = 0; ifov < 2; ifov++)
		{
		for (ior = 0; ior < 2; ior++)
			{
			TestViews[iv].idNum = iv;
			TestViews[iv].windowSize[0] = winSize[iwin][0];
			TestViews[iv].windowSize[1] = winSize[iwin][1];
			TestViews[iv].camera.fov = fov[ifov];
			TestViews[iv].camera.orientation[0]=orientation[ior][0];
			TestViews[iv].camera.orientation[1]=orientation[ior][1];
			TestViews[iv].camera.orientation[2]=orientation[ior][2];
			TestViews[iv].camera.orientation[3]=orientation[ior][3];

			TestViews[iv].camera.location[0] = 0;
			TestViews[iv].camera.location[1] = 0;
			TestViews[iv].camera.location[2] = 0;
			TestViews[iv].camera.near = 0.001;
			TestViews[iv].camera.far = 20.0;

			iv++;
			}
		}
	    }
#endif
	

	// generate points
	ustart (5019);  // seed for points

	for (i = 0; i < NumTestSegments; i++)
	    {
	    for (j = 0; j < 2; )
		{
		TestSegments[i][j][0] = -1.0 + 2 * uni ();
		TestSegments[i][j][1] = -1.0 + 2 * uni ();
		TestSegments[i][j][2] = -1.0 + 2 * uni ();

		if ( (  TestSegments[i][j][0] * TestSegments[i][j][0]  +
			TestSegments[i][j][1] * TestSegments[i][j][1]  +
			TestSegments[i][j][2] * TestSegments[i][j][2]  ) <= 1.0)
			{
			// We want only points inside the unit sphere
			j++;
			}

		}
	    }


#if 0

	TestSegments[0][0][0] = -1;
	TestSegments[0][0][1] =  0;
	TestSegments[0][0][2] = -1.1;
	TestSegments[0][1][0] =  1;
	TestSegments[0][1][1] =  0;
	TestSegments[0][1][2] = -1.1;

	// Here are some test segments that fail when used with
	// the hard-coded view given above.  
	// In all cases (except the one noted), they do not render at all.

	TestSegments[1][0][0] = 0.307583 ;
	TestSegments[1][0][1] = 0.184034 ;
	TestSegments[1][0][2] = -0.000713 ;
	TestSegments[1][1][0] = -0.045518 ;
	TestSegments[1][1][1] = -0.471818 ;
	TestSegments[1][1][2] = -0.338698 ;


	TestSegments[1][0][0] = 1;
	TestSegments[1][0][1] = 0;
	TestSegments[1][0][2] = 0 ;
	TestSegments[1][1][0] = 0;
	TestSegments[1][1][1] = -1;
	TestSegments[1][1][2] = -2;

	TestSegments[1][0][0] = 1;
	TestSegments[1][0][1] = 0;
	TestSegments[1][0][2] = -0.0005 ;
	TestSegments[1][1][0] = 0;
	TestSegments[1][1][1] = -1;
	TestSegments[1][1][2] = -2;

	// This one works
	TestSegments[1][0][0] = 1;
	TestSegments[1][0][1] = 0;
	TestSegments[1][0][2] = -0.001 ;
	TestSegments[1][1][0] = 0;
	TestSegments[1][1][1] = -1;
	TestSegments[1][1][2] = -2;

	TestSegments[1][0][0] = 1;
	TestSegments[1][0][1] = 0;
	TestSegments[1][0][2] = -0.0009 ;
	TestSegments[1][1][0] = 0;
	TestSegments[1][1][1] = -1;
	TestSegments[1][1][2] = -2;

	TestSegments[1][0][0] = 1;
	TestSegments[1][0][1] = 0;
	TestSegments[1][0][2] = -0.00099 ;
	TestSegments[1][1][0] = 0;
	TestSegments[1][1][1] = -1;
	TestSegments[1][1][2] = -2;

	//
	// Note that these fail when one of the points has z value
	// between zero and -near clipping plane.  
	//


	// works
	TestSegments[1][0][0] = 1;
	TestSegments[1][0][1] = 0;
	TestSegments[1][0][2] = 0.01 ;
	TestSegments[1][1][0] = 0;
	TestSegments[1][1][1] = -1;
	TestSegments[1][1][2] = -2;

	// fails
	TestSegments[1][0][0] = 1;
	TestSegments[1][0][1] = 0;
	TestSegments[1][0][2] = 0.001 ;
	TestSegments[1][1][0] = 0;
	TestSegments[1][1][1] = -1;
	TestSegments[1][1][2] = -2;

	TestSegments[1][0][0] = 1;
	TestSegments[1][0][1] = 0;
	TestSegments[1][0][2] = 0.0011 ;
	TestSegments[1][1][0] = 0;
	TestSegments[1][1][1] = -1;
	TestSegments[1][1][2] = -2;

	TestSegments[1][0][0] = 1;
	TestSegments[1][0][1] = 0;
	TestSegments[1][0][2] = 0.0015 ;
	TestSegments[1][1][0] = 0;
	TestSegments[1][1][1] = -1;
	TestSegments[1][1][2] = -2;

	TestSegments[1][0][0] = 1;
	TestSegments[1][0][1] = 0;
	TestSegments[1][0][2] = 0.00151 ;
	TestSegments[1][1][0] = 0;
	TestSegments[1][1][1] = -1;
	TestSegments[1][1][2] = -2;

	TestSegments[1][0][0] = 1;
	TestSegments[1][0][1] = 0;
	TestSegments[1][0][2] = 0.00152;
	TestSegments[1][1][0] = 0;
	TestSegments[1][1][1] = -1;
	TestSegments[1][1][2] = -2;

	// works
	TestSegments[1][0][0] = 1;
	TestSegments[1][0][1] = 0;
	TestSegments[1][0][2] = 0.00153;
	TestSegments[1][1][0] = 0;
	TestSegments[1][1][1] = -1;
	TestSegments[1][1][2] = -2;


	//
	// So actually it's more complicated than what I said above.
	// The near clipping plane is at 0.001 (z= -0.001) and we are observing
	// that failure occurs when one of the end points of the line segment
	// has a (camera coord) z value between +0.00152 and -0.001 .
	//

#if 0
	TestViews[0].windowSize[0] = 907;
	TestViews[0].windowSize[1] = 907;
	TestViews[0].camera.location[0] = 0;
	TestViews[0].camera.location[1] = 0;
	TestViews[0].camera.location[2] = 0;
	TestViews[0].camera.orientation[0] = 0;
	TestViews[0].camera.orientation[1] = 0;
	TestViews[0].camera.orientation[2] = 0;
	TestViews[0].camera.orientation[3] = 1;
	TestViews[0].camera.fov = 30.0;
	TestViews[0].camera.near = 0.001;
	TestViews[0].camera.far = 20.0;




	TestSegments[1][0][0] = 0.691774 ;
	TestSegments[1][0][1] = -0.680486 ;
	TestSegments[1][0][2] = 0.002896 ;
	TestSegments[1][1][0] = -0.230379 ;
	TestSegments[1][1][1] = 0.567138 ;
	TestSegments[1][1][2] = -0.781290 ;

	// works
	TestSegments[1][0][0] = 0.691774 ;
	TestSegments[1][0][1] = -0.680486 ;
	TestSegments[1][0][2] = 0.01;
	TestSegments[1][1][0] = -0.230379 ;
	TestSegments[1][1][1] = 0.567138 ;
	TestSegments[1][1][2] = -0.781290 ;

	TestSegments[1][0][0] = 0.691774 ;
	TestSegments[1][0][1] = -0.680486 ;
	TestSegments[1][0][2] = 0.0071;
	TestSegments[1][1][0] = -0.230379 ;
	TestSegments[1][1][1] = 0.567138 ;
	TestSegments[1][1][2] = -0.781290 ;
#endif





#endif






	}  // end of initRenderingTestCases


static int CurrentCase = 0;


static void
ungetLastRenderingTestCase ()
	{
	if (CurrentCase > 0)
		{
		CurrentCase--;
		}
	}

static int
getNextRenderingTestCase (View *view, Segment seg)
	{
	int currentView, currentSeg;

	if (CurrentCase >= (NumTestViews * NumTestSegments))
		{
		return -1;
		}



	currentView = CurrentCase / NumTestSegments;
// currentView = 0;
	currentSeg = CurrentCase % NumTestSegments;

	memcpy (view, TestViews+currentView, sizeof (View));
	memcpy (seg, TestSegments+currentSeg, sizeof (Segment));

	CurrentCase++;

#if 0
if (CurrentCase > 2000)
{
exit (-1);
}
#endif

	if ((CurrentCase % 1000) == 0)
		{
		fprintf (stderr, "At case %d\n", CurrentCase);
		}

	return 0;
	}  // end of getNextRenderingTestCase




////////////////////////////////////////////////////////////////

static void setupViewing (View *view)

	{
	GLdouble aspectRatio;
	double cameraMat[4][4];
	

	//
	// Doing this routine every time (rather than only when the
	// viewing changes) seems to degrade performance by only
	// about 7% on a 500x500 window (on acadie).
	//

	if ( (view->windowSize[0] > CurrentWindowSize[0]) ||
	     (view->windowSize[1] > CurrentWindowSize[1])     )
		{
		fprintf (stderr, "Specified viewport size (%d, %d) "
		  "is greater than current window size (%d, %d).\n",
		  view->windowSize[0], view->windowSize[1],
		  CurrentWindowSize[0], CurrentWindowSize[1]);
		exit (-1);
		}

#if 0
	CurrentViewportSize[0] = view->windowSize[0];
	CurrentViewportSize[1] = view->windowSize[1];
#endif

   	glViewport (0, 0, view->windowSize[0], view->windowSize[1] );

	// If this divide fails because windowSize[1] == 0, then you're getting
	// what you deserve.
	aspectRatio = 
	  ((double)view->windowSize[0]-1) / ((double)view->windowSize[1]-1);
	
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	gluPerspective (view->camera.fov, aspectRatio, 
				view->camera.near, view->camera.far);

	glMatrixMode (GL_MODELVIEW);

	glLoadIdentity ();

	quatToMat (view->camera.orientation, cameraMat);
	gluLookAt (
			view->camera.location[0],
			view->camera.location[1],
			view->camera.location[2],
			-cameraMat[2][0],
			-cameraMat[2][1],
			-cameraMat[2][2],
			cameraMat[1][0],
			cameraMat[1][1],
			cameraMat[1][2]   );


	}  // end of setupViewing

static unsigned char *Scrbuf = NULL;
static int ScrbufSize = 0;

static int 
getRenderedPixels  (
  View *view,
  Segment seg,
  int *numRenderedPixels, 
  int pixelXY[][2] ) 
	{
	unsigned char *pixel;
	int i, j;
	int newScrbufSize;

	newScrbufSize = 
	  4 * view->windowSize[0] * view->windowSize[1] * sizeof(unsigned char);

	if (ScrbufSize < newScrbufSize)
		{
		if (Scrbuf != NULL)
			{
			free (Scrbuf);
			}
		

		ScrbufSize = newScrbufSize;
		Scrbuf = (unsigned char *) malloc (ScrbufSize);

		if (Scrbuf == NULL)
			{
			fprintf (stderr, 
			"getRenderedPixels: Fatal error allocating memory.\n");
			exit (-1);
			}
		}


	setupViewing ( view );

	// Begin rendering

#if 1
	// Clear the frame buffer
	glClear (GL_COLOR_BUFFER_BIT);
#endif


	glColor3f (1.0, 1.0, 1.0);
	glBegin (GL_LINES);
	    GLVERTEX ( seg[0][0], seg[0][1], seg[0][2]);
	    GLVERTEX ( seg[1][0], seg[1][1], seg[1][2]);
	glEnd ();


	glFinish ();
	// End rendering


	// We've just rendered into the back buffer, so read the pixels
	// from there.
	// glReadBuffer (GL_BACK);  // or front?????


	glReadPixels (0, 0, view->windowSize[0], view->windowSize[1], 
				GL_RGBA, GL_UNSIGNED_BYTE, Scrbuf);

	*numRenderedPixels = 0;
	pixel = Scrbuf;
	for (j = 0; j < view->windowSize[1]; j++)
		{
		for (i = 0; i < view->windowSize[0]; i++)
			{
			if (pixel[0] != 0)
				{

				pixelXY[*numRenderedPixels][0] = i;
				pixelXY[*numRenderedPixels][1] = j;
				(*numRenderedPixels)++;
				}
			pixel += 4;
			}
		}

	return 0;
	}  // end of getRenderedPixels




static int
clipSegmentToPlane (double planeEq[4], Segment seg)
	{

	double denomin;
	double t;
	int replace;
	double d[2];

#if 0
if (Debug)
{
printf ("\nclipSegmentToPlane:\n");
printf ("   clipping against plane %f %f %f %f\n",
planeEq[0],
planeEq[1],
planeEq[2],
planeEq[3]);
printf ("     before segment = \n");
printSeg (seg);
}

#endif


	// array d gives signed distance from each seg end point to plane

	d[0] = 	planeEq[0] * seg[0][0] +
		planeEq[1] * seg[0][1] +
		planeEq[2] * seg[0][2] +
		planeEq[3];

	d[1] = 	planeEq[0] * seg[1][0] +
		planeEq[1] * seg[1][1] +
		planeEq[2] * seg[1][2] +
		planeEq[3];


	if ((d[0] >= 0) && (d[1] >= 0))
		{
		// the segment is entirely on the "interior" side of the plane 
		return 0;
		}
	else if ( (d[0] < 0) && (d[1] < 0) )
		{
		// the segment is entirely on the "exterior" side of the plane 
		return -1;
		}

	// Else the segment crosses the plane and we must find the
	// point of intersection and clip the segment

	// Line is seg[0] + t * (seg[1] - seg[0])
	// Plane is Ax + By + Cz + D = 0, with A,B,C,D from planeEq

	// Substitute line into plane equation and solve for t

	denomin = 
		planeEq[0] * (seg[0][0] - seg[1][0]) +
		planeEq[1] * (seg[0][1] - seg[1][1]) +
		planeEq[2] * (seg[0][2] - seg[1][2]) ;

	if (denomin == 0)
		{
		// then the segment is ON the plane, so we accept the entire seg
		return 0;
		}

	t = d[0] / denomin;

	if ( (t < 0) || (t > 1))
		{
		// Something is wrong: the point of intersection is not on 
		// the segment between the points.
		// We're going to exit.
		fprintf (stderr, "Error in segment / plane intersection.\n");
		exit (-1);
		}


	replace = (d[0] < 0) ? 0 : 1;
	seg[replace][0] = seg[0][0] + t * (seg[1][0] - seg[0][0]);
	seg[replace][1] = seg[0][1] + t * (seg[1][1] - seg[0][1]);
	seg[replace][2] = seg[0][2] + t * (seg[1][2] - seg[0][2]);

#if 1
if (Debug)
{
printf ("     clipped segment = \n");
printSeg (seg);
printf ("\n\n\n");
}
#endif

	return 0;
	}  // end of clipSegmentToPlane



static void
worldToCameraSeg (Camera *camera, Segment worldSeg, Segment cameraSeg)
	{
	double tmpPt[3];
	double camToWorldRot[4][4];
	double worldToCamRot[4][4];
	int i;

	// Not very efficient, but who cares?

	quatToMat (camera->orientation, camToWorldRot);
	transposeMat (camToWorldRot, worldToCamRot);

	for (i = 0; i < 2; i++)
		{
		tmpPt[0] = worldSeg[i][0] - camera->location[0];
		tmpPt[1] = worldSeg[i][1] - camera->location[1];
		tmpPt[2] = worldSeg[i][2] - camera->location[2];
		xformPointByMat (tmpPt, worldToCamRot, cameraSeg[i]);
		}

	}  // end of worldToCameraSeg


static void
cameraToWorldSeg (Camera *camera, Segment cameraSeg, Segment worldSeg)
	{
	double tmpPt[3];
	double camToWorldRot[4][4];
	int i;

	// Not very efficient, but who cares?

	quatToMat (camera->orientation, camToWorldRot);

	for (i = 0; i < 2; i++)
		{
		xformPointByMat (cameraSeg[i], camToWorldRot, tmpPt);
		worldSeg[i][0] = tmpPt[0] + camera->location[0];
		worldSeg[i][1] = tmpPt[1] + camera->location[1];
		worldSeg[i][2] = tmpPt[2] + camera->location[2];
		}

	}  // end of cameraToWorldSeg




static void
cameraToPixelSeg (
  View *view, 
  Segment cameraSeg, 
  Segment pixelSeg
  )
	{
	double botEdge, vertExtent, aspectRatio, horizExtent, leftEdge;
	int i;

	double projectedXY[2];




	botEdge = - tan (DEG_TO_RAD*view->camera.fov/2.0);
	vertExtent = 2 * tan (DEG_TO_RAD*view->camera.fov/2.0);




	aspectRatio = ((double)(view->windowSize[0]-1)) / 
				((double)(view->windowSize[1]-1));
	horizExtent = vertExtent * aspectRatio;
	leftEdge = botEdge * aspectRatio;


	for (i = 0; i < 2; i++)
		{
		projectedXY[0] = - cameraSeg[i][0] / cameraSeg[i][2];
		projectedXY[1] = - cameraSeg[i][1] / cameraSeg[i][2];

		pixelSeg[i][0] = -0.5 + ((projectedXY[0] - leftEdge) / 
					horizExtent) * (view->windowSize[0]);
		pixelSeg[i][1] = -0.5 + ((projectedXY[1] - botEdge) / 
					vertExtent) * (view->windowSize[1]);
		}

	}  // end of cameraToPixelSeg



static int clipToViewingFrustum ( 
  View *view, 
  Segment seg, 
  Segment clippedSeg,
  Segment clippedSegPixelCoords
  )
	{
	

	// The viewing frustum is defined by six planes.  We will find
	// the intersection of the segment with each of these planes.
	//
	// We are going to transform the points into "camera" coordinates, 
	// calculate the intersections, then transform back.  This makes
	// the intersection calculations a whole lot easier.

	Segment cameraSeg;
	double planeEq[6][4];
	double aspectRatio, horizFOV_o2;
	int i;


	worldToCameraSeg (&(view->camera), seg, cameraSeg);


	// We construct the plane equations for the six planes in camera coords.
	// These will be the four coefficients of the equation 
	//   Ax + By + Cz + D = 0 for each of the six planes.
	// The plane equation coefficients will be stated so that the left
	// side of the equation will be positive when an (x,y,z) point is
	// in the "inside" half-space relative to the plane.

	// Near plane
	planeEq[0][0] = 0;
	planeEq[0][1] = 0;
	planeEq[0][2] = - 1;
	planeEq[0][3] = - view->camera.near;

	// Far plane
	planeEq[1][0] = 0;
	planeEq[1][1] = 0;
	planeEq[1][2] = 1;
	planeEq[1][3] = view->camera.far;

	// All of the other planes of the frustum go through the origin
	// which means that D = 0 in the plane equation.

	// The angular field of view specifies the vertical field
	// of view for the window.  

	// Upper plane
	//
	// The unit normal to the upper plane is  (0, sin(-fov/2), cos(-fov/2))
	planeEq[2][0] = 0;
	planeEq[2][1] = - cos (DEG_TO_RAD*view->camera.fov/2.0);
	planeEq[2][2] = - sin (DEG_TO_RAD*view->camera.fov/2.0);
	planeEq[2][3] = 0;

	// Lower plane
	planeEq[3][0] = 0;
	planeEq[3][1] =   cos (DEG_TO_RAD*view->camera.fov/2.0);
	planeEq[3][2] = - sin (DEG_TO_RAD*view->camera.fov/2.0);
	planeEq[3][3] = 0;


	// So now we get the horizontal field of view (actually the
	// half angle.
	aspectRatio = ((double)(view->windowSize[0]-1)) / 
				((double)(view->windowSize[1]-1));
	horizFOV_o2 =  
		atan (tan (DEG_TO_RAD*view->camera.fov/2.0) * aspectRatio ) ;

	// Right plane
	planeEq[4][0] = - cos (horizFOV_o2);
	planeEq[4][1] = 0;
	planeEq[4][2] = - sin (horizFOV_o2);
	planeEq[4][3] = 0;

	// Left plane
	planeEq[5][0] =   cos (horizFOV_o2);
	planeEq[5][1] = 0;
	planeEq[5][2] = - sin (horizFOV_o2);
	planeEq[5][3] = 0;


	for (i = 0; i < 6; i++)
		{
		if ((clipSegmentToPlane (planeEq[i], cameraSeg)))
			{
			return -1;
			}
		}


	cameraToPixelSeg (view, cameraSeg, clippedSegPixelCoords);
	cameraToWorldSeg (&(view->camera), cameraSeg, clippedSeg);
	return 0;
	}  // end of clipToViewingFrustum

////////////////////////////////////////////////////////////////////////////////////







////////////////////////////////////////////////////////////////////////////////////

/*

For each rendered pixel, calculate the distance from the line
to the pixel, and the distance along the line segment relative
to the first point on the line segment.

Note that the distance along the segment increases for
pixels closer to the second end point.  

Nominally, you would expect the "along" distance to range 
from zero to the length of the line segment. Deviations from
this represent undershoot and overshoot.

*/

static int
calcLinePointDistances (
  Segment clippedSegPixelCoords, 
  int numRenderedPixels, 
  int renderedPixels[][2], 
  double signedDistPixelToLine[], 
  double distAlongLine[]
  )
	{

	// line equation ax + by + c = 0, where (a, b) is a unit vector
	double seg_a, seg_b, seg_c;  
	double perp_a, perp_b, perp_c;  
	double len;

	int i;



	seg_a = clippedSegPixelCoords[1][1] - clippedSegPixelCoords[0][1];
	seg_b = clippedSegPixelCoords[0][0] - clippedSegPixelCoords[1][0];

	len = sqrt (seg_a*seg_a + seg_b*seg_b);
	seg_a /= len;
	seg_b /= len;

	seg_c = - (seg_a*clippedSegPixelCoords[0][0] + 
				seg_b*clippedSegPixelCoords[0][1]);

	perp_a =  seg_b;
	perp_b = -seg_a;


	for (i = 0; i < numRenderedPixels; i++)
		{
		perp_c = - (perp_a*renderedPixels[i][0] + 
				perp_b*renderedPixels[i][1]);

		signedDistPixelToLine[i] = 
			seg_a * renderedPixels[i][0] + 
			seg_b * renderedPixels[i][1] + 
			seg_c;


		distAlongLine[i] = 
			perp_a * clippedSegPixelCoords[0][0] + 
			perp_b * clippedSegPixelCoords[0][1] + 
			perp_c;

#if 0
if ((i==0) || (i == numRenderedPixels-1))
{
printf ("pixel[%d] = %d %d    to %f   along %f\n", 
i, renderedPixels[i][0], renderedPixels[i][1],
signedDistPixelToLine[i], distAlongLine[i]);
}
#endif

		}
	

	return 0;
	}  // end of calcLinePointDistances

typedef struct line_distance
	{
	double alongLine;
	double toLine;
	int index;
	} LineDistance;

static int compareLineDistances (const void *a, const void *b)
	{

	return  
		((LineDistance *)a)->alongLine
		<
		((LineDistance *)b)->alongLine ? -1:1;
		;
	}  // end of compareLineDistances

/////////////////////////////////////////////////////////////////////


// In order to calculate overshoot and undershoot errors, we need
// to solve a simple integral.
//
// The following function, integrateSegPtDistance, computes the
// integral, over the points of a given segment, of the distance to
// a given point.  A closed-form solution is used.



/*

First we transform the problem to the case where the segment
lies on the X axis.  Then we integrate, with respect ot x, the distance from
(x,0) to the (transformed) point as a function of x.

For the closed form solution, see CRC Standard Mathematical Tables
17th Edition p 424 and 421.

We use formulae 242 and 238:

  Let the point be (x1, y1)

  The distance from a point (x,0) to that point is:

  f(x) = sqrt ( (x-x1)^2 + y1^2)

       = sqrt ( x^2 - 2*x1*x + (x1^2 + y1^2) )

  So we need 

    integral f(x) dx

  In order to conform to formula 242, 
  let c = 1, b = -2*x1  a = x1^2 + y1^2.

  And q = 4ac - b^2 = 4x1^2 + 4y1^2 - 4x1^2 = 4y1^2
  and k = 4c/q = 1/y1^2

  If q != 0, then by formula 242:

    integral f(x) dx = ( (x - x1) * f(x) / 2 ) + (y1^2/2)*integral (1/f(x))dx

     by formula 237

		= ( (x - x1) * f(x) / 2 ) + (y1^2/2)* log (2*f(x) + 2x - 2x1)

		= ( (x - x1) * f(x) / 2 ) + 
		      (y1^2/2)*log (f(x) + x - x1) + y1^2/2*log(2)

    note that the last term is a constant, so it can be dropped.

    This corresponds with the results given by 

	http://integrals.wolfram.com/index.jsp


  When q is 0: Note that q==0 iff y1 == 0 and in this case, the system 
  reduces to integrating |x-x1|.  This integral is simply 

		(x - x1) * |x - x1| / 2

*/

//
// Here's an implementation of the closed form solution
//
double integrateSegPtDistance (double seg[2][2], double pt[2])
	{
	double fOf0, fOfSegLen, integral0, integralSegLen, segLen,
		x1, y1, alongLine, fromLine, dir[2], sqD;


	// This is a bit wordier than necessary, but I hope that
	// it helps to clarify things.

	// First, transform system so that the segment is along the X axis.

	// We are going to transform so that seg[0][*] is at the
	// origin and seg[1][*] is on the positive X axis.

	dir[0] = seg[1][0] - seg[0][0];
	dir[1] = seg[1][1] - seg[0][1];

	segLen = sqrt ( dir[0]*dir[0] + dir[1]*dir[1] );

	dir[0] /= segLen;
	dir[1] /= segLen;

	// So dir is the 2D direction vector of the line segment

	// dot product of dir vector & vector to pt gives distance
	// along line.
	alongLine = 	dir[0] * (pt[0] - seg[0][0]) +
			dir[1] * (pt[1] - seg[0][1]) ;


	// This is the square of the distance from pt to seg[0]
	sqD = 
                ( (pt[0] - seg[0][0])*(pt[0] - seg[0][0]) +
                  (pt[1] - seg[0][1])*(pt[1] - seg[0][1]) )

                -

                // This is the square of the dist from seg[0][*] to projection
                // of pt to line
                alongLine * alongLine;

	fromLine = (sqD > 0) ? sqrt (sqD) : 0.0;

	x1 = alongLine;
	y1 = fromLine;


	if (y1 == 0)
		{
		integral0 = - x1 * fabs (x1) / 2.0;
		integralSegLen = (segLen - x1) * fabs (segLen - x1) / 2.0;
		}
	else
		{


		// So we will integrate the distance between 
		//  (x,0) and (x1,y1) as x goes from 0 to segLen.

		fOf0 = sqrt ( x1*x1 + y1*y1);
		fOfSegLen = sqrt ( (segLen - x1) * (segLen - x1) + y1*y1 );

		integral0 = (-x1*fOf0/2.0) + (y1*y1/2.0)*log (fOf0 - x1);
		integralSegLen = 
			((segLen - x1)*fOfSegLen/2.0) +
			(y1*y1/2.0)*log (fOfSegLen + (segLen - x1));

		}

	return integralSegLen - integral0;
	}  // end of integrateSegPtDistanceB





//
// Here's an implementation that approximates the integral's solution
// based on sampling the segment at 1000 points.  We use this to check 
// the general correctness of the closed form solution.
//
double integrateSegPtDistanceCheck (double seg[2][2], double pt[2])
	{
	double result;
	double len;
	double t, segPt[2], d; 
	int i;


	// The length of each segment:
	len = sqrt (
	  ((seg[0][0] - seg[1][0]) * (seg[0][0] - seg[1][0])) +
	  ((seg[0][1] - seg[1][1]) * (seg[0][1] - seg[1][1])) ) / 1000.0;

	result = 0.0;

	for (i = 0; i < 1000; i++)
		{
		t = ((double)i)/1000.0;
		segPt[0] = t*seg[0][0] + (1.0 - t) * seg[1][0];
		segPt[1] = t*seg[0][1] + (1.0 - t) * seg[1][1];

		d = sqrt 
			(
			((segPt[0] - pt[0]) * (segPt[0] - pt[0])) +
			((segPt[1] - pt[1]) * (segPt[1] - pt[1]))
			);

		result += d*len;
		}

	return result;

	}  // end of integrateSegPtDistanceA




#if 1

/*
This routine serves calcOvershoot by dividing up the overshoot
segment into two sections if it crosses the x axis. It also 
tells on which side of the X axis each subsection lies.
*/
static void
divideOvershootSeg (	double seg[2][2], 
			int *nSubSeg, 
			double subSeg[2][2][2], 
			int signs[2])

	{
	double p, x;

	if ((seg[0][1]*seg[1][1]) >= 0)
		{
		// The two segment end points are on the same side of the
		// x axis, so no subdivision is needed.
		*nSubSeg = 1;
		subSeg[0][0][0] = seg[0][0];
		subSeg[0][0][1] = seg[0][1];
		subSeg[0][1][0] = seg[1][0];
		subSeg[0][1][1] = seg[1][1];
		signs[0] = ((seg[0][1] < 0) || (seg[1][1] < 0)) ? -1 : 1;
		}
	else
		{
		// The two segment end points are on opposite sides of the
		// x axis, so we have to split the segment into two
		// subsegments.

		*nSubSeg = 2;

		// first point of first seg
		subSeg[0][0][0] = seg[0][0];
		subSeg[0][0][1] = seg[0][1];

		// second point of second seg
		subSeg[1][1][0] = seg[1][0];
		subSeg[1][1][1] = seg[1][1];

		// The middle point has y = 0
		subSeg[0][1][1] = 0;
		subSeg[1][0][1] = 0;

		p = seg[0][1] / (seg[0][1] - seg[1][1]);
		
		x = seg[0][0] + p * (seg[1][0] - seg[0][0]);
		subSeg[0][1][0] = x;
		subSeg[1][0][0] = x;
		signs[0] = (seg[0][1] >= 0) ? 1 : -1;
		signs[1] = (seg[1][1] >= 0) ? 1 : -1;
		}

	return;
	}  // end of divideOvershootSeg


//
// The following routine calculates overshoot error associated with 
// the segment defined by the pixel segment from pixDist0 and pixDist1
// relative to the true segement (0,segLen)
//
static void
calcOvershoot (double segLen, LineDistance *pixDist0, LineDistance *pixDist1,  
		double *overshootErr, double *signedOvershootErr)
	{
	double pt[2];
	double seg[2][2];
	double subSeg[2][2][2];
	int nSubSeg;
	int signs[2];
	double p;
	double err;
	int i;

#if 0
printf ("BEGIN overshoot: segLen = %f   pix0 %f %f  pix1 %f %f  \n",
segLen,
pixDist0->alongLine,
pixDist0->toLine,
pixDist1->alongLine,
pixDist1->toLine
);
#endif

	*overshootErr = 0;
	*signedOvershootErr = 0;

	// We look to see if the "alongLine" values of the pixel segment 
	// lie entirely within the range that corresponds to the interior 
	// of the true clipped segment, which extends from 0 to segLen.
	// If it does not, then we have an overshoot situation.  
	// We calculate an error for the part that is not within the true
	// seg range by computing the integral over the overshoot part of
	// the segment of the distance to the appropriate true clipped seg
	// end point.

	// pixDist0 and pixDist1 are modified to reflect the portion 
	// of the pixel segment that is within the range of the 
	// true segment.

		if (pixDist0->alongLine < 0)
			{
			// overshoot on first end

			pt[0] = pt[1] = 0;
			seg[0][0] = pixDist0->alongLine;
			seg[0][1] = pixDist0->toLine;

			if (pixDist1->alongLine <= 0)
				{
				// entire pixel segment is out of range
				seg[1][0] = pixDist1->alongLine;
				seg[1][1] = pixDist1->toLine;
				// update pixDist0 to be pixDist1
				pixDist0->alongLine = pixDist1->alongLine;
				pixDist0->toLine = pixDist1->toLine;
				}
			else
				{
				// only a portion is out of range
				seg[1][0] = 0;
				p = (- pixDist0->alongLine) /
			         (pixDist1->alongLine - pixDist0->alongLine);
				seg[1][1] = pixDist0->toLine + 
				      p * (pixDist1->toLine - pixDist0->toLine);
				// update pixDist0 to be pixDist1
				pixDist0->alongLine = 0;
				pixDist0->toLine = seg[1][1];
				}

			divideOvershootSeg (seg, &nSubSeg, subSeg, signs);
			for (i = 0; i < nSubSeg; i++)
				{
				err = integrateSegPtDistance (subSeg[i], pt);
				*overshootErr += err;
				*signedOvershootErr += err*signs[i];
				}
			}

		if (pixDist1->alongLine > segLen)
			{
			// overshoot on second end
			pt[0] = segLen;
			pt[1] = 0;
			seg[1][0] = pixDist1->alongLine;
			seg[1][1] = pixDist1->toLine;
			if (pixDist1->alongLine >= segLen)
				{
				// entire pixel segment is out of range
				seg[0][0] = pixDist0->alongLine;
				seg[0][1] = pixDist0->toLine;
				// update pixDist0 to be pixDist1
				pixDist1->alongLine = pixDist0->alongLine;
				pixDist1->toLine = pixDist0->toLine;
				}
			else
				{
				// only a portion is out of range
				seg[0][0] = segLen;
				p = (segLen - pixDist0->alongLine) /
				    (pixDist1->alongLine - pixDist0->alongLine);
				seg[0][1] = pixDist0->toLine + 
				      p * (pixDist1->toLine - pixDist0->toLine);
				pixDist1->alongLine = segLen;
				pixDist1->toLine = seg[0][1];
				}

			divideOvershootSeg (seg, &nSubSeg, subSeg, signs);
			for (i = 0; i < nSubSeg; i++)
				{
				err = integrateSegPtDistance (subSeg[i], pt);
				*overshootErr += err;
				*signedOvershootErr += err*signs[i];
				}

			}
		
#if 0
printf ("END overshoot: segLen = %f   pix0 %f %f  pix1 %f %f  err %f %f\n\n",
segLen,
pixDist0->alongLine,
pixDist0->toLine,
pixDist1->alongLine,
pixDist1->toLine,
*overshootErr,
*signedOvershootErr);
#endif

	}  // end of calcOvershoot










#else


//
// The following routine calculates overshoot error associated with 
// the segment defined by the pixel segment from pixDist0 and pixDist1
// relative to the true segement (0,segLen)
//
static void
calcOvershoot (double segLen, LineDistance *pixDist0, LineDistance *pixDist1,  
		double *overshootError)
	{
	double pt[2];
	double seg[2][2];
	double p;

#if 0
printf ("BEGIN overshoot: segLen = %f   pix0 %f %f  pix1 %f %f  err %f\n",
segLen,
pixDist0->alongLine,
pixDist0->toLine,
pixDist1->alongLine,
pixDist1->toLine,
*lineError);
#endif

	*overshootError = 0;

	// We look to see if the "alongLine" values of the pixel segment 
	// lie entirely within the range that corresponds to the interior 
	// of the true clipped segment, which extends from 0 to segLen.
	// If it does not, then we have an overshoot situation.  
	// We calculate an error for the part that is not within the true
	// seg range by computing the integral over the overshoot part of
	// the segment of the distance to the appropriate true clipped seg
	// end point.

	// pixDist0 and pixDist1 are modified to reflect the portion 
	// of the pixel segment that is within the range of the 
	// true segment.

		if (pixDist0->alongLine < 0)
			{
			// overshoot on first end

			pt[0] = pt[1] = 0;
			seg[0][0] = pixDist0->alongLine;
			seg[0][1] = pixDist0->toLine;

			if (pixDist1->alongLine <= 0)
				{
				// entire pixel segment is out of range
				seg[1][0] = pixDist1->alongLine;
				seg[1][1] = pixDist1->toLine;
				// update pixDist0 to be pixDist1
				pixDist0->alongLine = pixDist1->alongLine;
				pixDist0->toLine = pixDist1->toLine;
				}
			else
				{
				// only a portion is out of range
				seg[1][0] = 0;
				p = (- pixDist0->alongLine) /
			         (pixDist1->alongLine - pixDist0->alongLine);
				seg[1][1] = pixDist0->toLine + 
				      p * (pixDist1->toLine - pixDist0->toLine);
				// update pixDist0 to be pixDist1
				pixDist0->alongLine = 0;
				pixDist0->toLine = seg[1][1];
				}

			*overshootError += integrateSegPtDistance (seg, pt);
			}

		if (pixDist1->alongLine > segLen)
			{
			// overshoot on second end
			pt[0] = segLen;
			pt[1] = 0;
			seg[1][0] = pixDist1->alongLine;
			seg[1][1] = pixDist1->toLine;
			if (pixDist1->alongLine >= segLen)
				{
				// entire pixel segment is out of range
				seg[0][0] = pixDist0->alongLine;
				seg[0][1] = pixDist0->toLine;
				// update pixDist0 to be pixDist1
				pixDist1->alongLine = pixDist0->alongLine;
				pixDist1->toLine = pixDist0->toLine;
				}
			else
				{
				// only a portion is out of range
				seg[0][0] = segLen;
				p = (segLen - pixDist0->alongLine) /
				    (pixDist1->alongLine - pixDist0->alongLine);
				seg[0][1] = pixDist0->toLine + 
				      p * (pixDist1->toLine - pixDist0->toLine);
				pixDist1->alongLine = segLen;
				pixDist1->toLine = seg[0][1];
				}


			*overshootError += integrateSegPtDistance (seg, pt);
			}
		
#if 0
printf ("END overshoot: segLen = %f   pix0 %f %f  pix1 %f %f  err %f\n\n",
segLen,
pixDist0->alongLine,
pixDist0->toLine,
pixDist1->alongLine,
pixDist1->toLine,
*overshootError);
#endif

	}  // end of calcOvershoot










#endif





//
// The following routine calculates undershoot error for a rendered
// series of pixels that extends from firstPix to lastPix relative
// to a true segment that extends from 0 to segLen.
//
static void
calcUndershoot (
  double segLen, 
  LineDistance *firstPix, 
  LineDistance *lastPix, 
  double *undershootErr,
  double *signedUndershootErr)
	{
	double seg[2][2], pt[2], err;

	// Undershoot error is calculated by integrating over
	// the portion the true segment that is not covered by
	// the rendered pixels.  We integrate the distance from
	// the this segment portion to the appropriate rendered
	// pixel end point.


	*undershootErr = 0;
	*signedUndershootErr = 0;

	// Is there undershoot at first end?
	if (firstPix->alongLine > 0)
		{
		// handle undershoot at first end
		seg[0][0] = 0;
		seg[0][1] = 0;
		seg[1][0] = firstPix->alongLine;
		seg[1][1] = 0;
		pt[0] = firstPix->alongLine;
		pt[1] = firstPix->toLine;
		err = integrateSegPtDistance (seg, pt);
		*undershootErr += err;
		*signedUndershootErr += (firstPix->toLine >= 0) ? err : -err;
		}

	if (lastPix->alongLine < segLen)
		{
		// handle undershoot at second end
		seg[0][0] = lastPix->alongLine;
		seg[0][1] = 0;
		seg[1][0] = segLen;
		seg[1][1] = 0;
		pt[0] = lastPix->alongLine;
		pt[1] = lastPix->toLine;
		err = integrateSegPtDistance (seg, pt);
		*undershootErr += err;
		*signedUndershootErr += (lastPix->toLine >= 0) ? err : -err;
		}

	} // end of calcUndershoot


	
static void
doChecks 
  (
  Segment clippedSegPixelCoords, 
  int numRenderedPixels, 
  int renderedPixels[][2],
  LineDistance lineDist[] 
  )
	{
	// This section contains to checks
	int j, endPt[2];
	double mn, mx, diff;
	int overshoot;

#if 0
	int k;
	int direction;
	double endPtDist[2];

	printf ("\nHere is the clipped segment in pixel coords:\n");
	printSeg (clippedSegPixelCoords);
	
	printf ("Here is the set of rendered pixels:\n");
	for (i = 0; i < numRenderedPixels; i++)
		{
		printf ("%d  index %d  along %f   pt %d %d\n",
			i,
			lineDist[i].index,
			lineDist[i].alongLine,
			renderedPixels[lineDist[i].index][0],
			renderedPixels[lineDist[i].index][1]);
		}


	direction = (lineDist[0].index < 
			lineDist[numRenderedPixels-1].index) ? 1 : -1;

	// Check whether the sorting makes a difference 
	for (i = 1; i < numRenderedPixels; i++)
		{
		if (lineDist[i].index != (lineDist[i-1].index+direction))
			{
			printf ("Pixels rearranged!\n");
			break;
			}
		}
#endif

	// check for over/under shoot of drawn line segment

	if (fabs(lineDist[0].alongLine) < 
		       fabs(lineDist[numRenderedPixels-1].alongLine))
		{
		endPt[0] = 0;
		endPt[1] = numRenderedPixels - 1;
		}
	else
		{
		endPt[1] = 0;
		endPt[0] = numRenderedPixels - 1;
		}

#if 0

	for (j = 0; j < 2; j++)
		{
		double len, d;
		len = 0;
		for (k = 0; k < 2; k++)
			{
			d = clippedSegPixelCoords[j][k] - 
				renderedPixels[lineDist[endPt[j]].index][k];
			len += d*d;
			}
		endPtDist[j] = sqrt (len);

		printf ("endPtDist ( %d, %d ) to ( %f, %f) is %f ",
			renderedPixels[lineDist[endPt[j]].index][0],
			renderedPixels[lineDist[endPt[j]].index][1],
			clippedSegPixelCoords[j][0],
			clippedSegPixelCoords[j][1],
			endPtDist[j]);
		if (endPtDist[j] > 1.0)
			{
			printf (" ERROR !?!?!?!?!");
			}
		if ( 
			(renderedPixels[lineDist[endPt[j]].index][0] == 0) |
			(renderedPixels[lineDist[endPt[j]].index][0] == 
						CurrentViewportSize[0]-1) |
			(renderedPixels[lineDist[endPt[j]].index][1] == 0) |
			(renderedPixels[lineDist[endPt[j]].index][1] == 
						CurrentViewportSize[1]-1) )
			{
			printf (" EDGE");
			}

		printf ("\n");


		}
#endif

	// Now let's look at end point over/under shoot
	mn = 0;
	mx = sqrt (
		(clippedSegPixelCoords[0][0] - clippedSegPixelCoords[1][0])*
		(clippedSegPixelCoords[0][0] - clippedSegPixelCoords[1][0])
		+
		(clippedSegPixelCoords[0][1] - clippedSegPixelCoords[1][1])*
		(clippedSegPixelCoords[0][1] - clippedSegPixelCoords[1][1])
		);

	overshoot = 0;
	for (j = 0; j < 2; j++)
		{
		printf ("endPt %d ( %d %d )", j, 
			renderedPixels[lineDist[endPt[j]].index][0],
			renderedPixels[lineDist[endPt[j]].index][1] );

		if (fabs(lineDist[endPt[j]].alongLine) < mn) 
			{
			overshoot = 1;
			printf (" over");
			diff = fabs (lineDist[endPt[j]].alongLine);
			}
		else if (fabs(lineDist[endPt[j]].alongLine) > mx)
			{
			overshoot = 1;
			printf (" over");
			diff = fabs (lineDist[endPt[j]].alongLine) - mx;
			}
		else
			{
			printf (" under");
			diff = MIN (
				(fabs (lineDist[endPt[j]].alongLine) - mn),
				(mx - fabs (lineDist[endPt[j]].alongLine))
				);
			}

		// printf (" %f %s\n", diff, (diff > 0.5)? "ERROR !?!?":"");
		printf (" %f %s\n", diff, (diff > 1.0)? "ERROR !?!?":"");

		if (overshoot)
			{
			int ii;
			printf ("shoot seg %f %f   %f %f    diff %f  \n",
				clippedSegPixelCoords[0][0], 
				clippedSegPixelCoords[0][1], 
				clippedSegPixelCoords[1][0], 
				clippedSegPixelCoords[1][1], diff);
			for (ii = 0; ii < numRenderedPixels; ii++)
				{
				printf ("shoot pix  %d %d %f %f\n",
					renderedPixels[lineDist[ii].index][0],
					renderedPixels[lineDist[ii].index][1],
					lineDist[ii].alongLine,
					lineDist[ii].toLine);
				}
			}
		}




#if 0
	printf ("end point distances %f %f", endPtDist[0], endPtDist[1]);
	if ( (endPtDist[0] > 1.0) || (endPtDist[1] > 1.0) )
		{
		printf (" ERROR !?!?!?!");
		}
	printf ("\n");
#endif


	} // end of check section



double dist2D_i_d (int a[2], double b[2])
	{
	return sqrt (
			(a[0] - b[0]) * (a[0] - b[0]) +
			(a[1] - b[1]) * (a[1] - b[1])
	            );
	}  // end of dist2D_i_d






static int 
calculateErrors (
  Segment clippedSegPixelCoords, 
  int numRenderedPixels, 
  int renderedPixels[][2], 
  double *signedPtError, 
  double *ptError, 
  double *signedLineError,
  double *lineError )
	{

	static double signedDistPixelToLine[MAX_NUM_RENDERED_PIXELS];
	static double distAlongLine[MAX_NUM_RENDERED_PIXELS];
	double signedArea, unsignedArea, signedTri0, signedTri1;
	static LineDistance lineDist[MAX_NUM_RENDERED_PIXELS];
	int i;
	double dist;
	double firstAlongDist, lastAlongDist;
	// double firstToDist, lastToDist;

#if 1
	double overshootError, signedOvershootError, 
		undershootError, signedUndershootError, segLen;

	// The length of the true clipped segment
	segLen = sqrt (
		(clippedSegPixelCoords[0][0] - clippedSegPixelCoords[1][0])*
		(clippedSegPixelCoords[0][0] - clippedSegPixelCoords[1][0])
		+
		(clippedSegPixelCoords[0][1] - clippedSegPixelCoords[1][1])*
		(clippedSegPixelCoords[0][1] - clippedSegPixelCoords[1][1])
		);
#endif


	calcLinePointDistances (
		clippedSegPixelCoords, 
		numRenderedPixels, 
		renderedPixels, 
		signedDistPixelToLine, 
		distAlongLine
		);

	// Note that these distances (distAlongLine and signedDistPixelToLine)
	// enable us to consider the geometry of the true line segment
	// and the rendered pixels in a much simpler configuration.
	//
	// The true clipped segment can be regarded as being on the
	// X axis extending from (0,0) to (segment length, 0).
	// The pixel X and Y coordinates correspond to the alongLine 
	// and signed toLine distances.
	//
	// In effect this is a simple rigid rotation and translation
	// of the original coordinate system.  It makes the subsequent
	// calculations much simpler.

	

	// For the point error metrics, I handle overshoot and undershoot
	// like this:
	//
	//	calc distance ALONG LINE from last pixel to the seg end point
	//	add that dist to ptError
	//	(DO NOT add anything to signedPtError)
	//
	//      so for undershoot, this just adds in the length that the 
	// 	true line segment sticks out beyond the length covered 
	// 	by the pixels.
	//
	//      And for overshoot, this adds in the length that the 
	// 	true line segment would have to be extended to match
	// 	the pixels.
	//
	//      We don't try to account for this in the signedPtError because
	//	the purpose of that metric is to measure bias to one side or
	//	the other of the segment, and we don't regard undershoot as
	//	having a bias.
	// 
	// Note that we add in the distance from each pixel to the line
	// containing the segment, whether the pixel is in an overshoot
	// condition, undershoot condition, or is within the segment.
	//

	firstAlongDist = lastAlongDist = distAlongLine[0];
	// firstToDist = lastToDist = signedDistPixelToLine[0];

	*signedPtError = *ptError = 0;
	for (i = 0; i < numRenderedPixels; i++)
		{
		// add in dist to line for all pixels
		*signedPtError += signedDistPixelToLine[i];
		*ptError += fabs (signedDistPixelToLine[i]);

		if (distAlongLine[i] < 0)
			{
			// overshoot at first end
			firstAlongDist = distAlongLine[i];
			dist =  - distAlongLine[i];
			*ptError += dist;
			}
		else if (distAlongLine[i] > segLen)
			{
			// overshoot at second end
			lastAlongDist = distAlongLine[i];
			dist = distAlongLine[i] - segLen;
			*ptError += dist;
			}
		else
			{
			// we're within the line segment

			if (distAlongLine[i] < firstAlongDist)
				{
				firstAlongDist = distAlongLine[i];
				// firstToDist = signedDistPixelToLine[i];
				}
			else if (distAlongLine[i] > lastAlongDist)
				{
				lastAlongDist = distAlongLine[i];
				// lastToDist = signedDistPixelToLine[i];
				}
			}
		}  // end of loop over pixels

	// Test for undershoot at the first end
	if (firstAlongDist > 0)
		{
		dist = firstAlongDist;
		*ptError += dist;
		}

	// Test for undershoot at the second end
	if (lastAlongDist < segLen)
		{
		dist = segLen - lastAlongDist;
		*ptError += dist;
		}


// printf ("------\n\n");




	*signedLineError = *lineError = 0;
	if (numRenderedPixels <= 1)
		{
		return 0;
		}

	// JGH: OK, I should have planned this out more carefully.  
	//      I need to traverse the points in order as determined
	//	by the distAlongLine.  So we need to sort them on 
	//      that basis.  For now, I'm just going to copy the
	//      needed information into an array of structures that
	//      I can hand to qsort.  To be more efficient, I should
	//      have placed them into such a structure to begin with.
	//
	//      
	for (i = 0; i < numRenderedPixels; i++)
		{
		lineDist[i].alongLine = distAlongLine[i];
		lineDist[i].toLine = signedDistPixelToLine[i];
		lineDist[i].index = i;
		}

	qsort (lineDist, numRenderedPixels, 
		sizeof (LineDistance), compareLineDistances);

	// the lineDist array is now in ascending order based on
	// pixel position alongLine




//////////////////////////////////////////////

if (0)
{
doChecks (clippedSegPixelCoords, numRenderedPixels, renderedPixels, lineDist);
}
//////////////////////////////////////////////


	//
	// There are three types of error that we accumulate:
	//
	//    - areas between drawn polyline and the true line segment
	//
	//    - overshoot errors 
	//		(when drawn segment extends beyond true segment)
	//
	//    - undershoot errors 
	//		(when true segment extends beyond drawn segment)
	//
	// Note that we include the overshoot and undershoot errors 
	// in lineError, but not in signedLineError.
	//


	// In terms of the "alongLine" coordinate, the true clipped
	// segment extends from 0 to segLen.

	for (i = 1; i < numRenderedPixels; i++)
		{
		LineDistance pixDist0, pixDist1;
		

		// We transfer the LineDistance structures to local
		// variables because we might need to change the
		// contents of the structures.

		memcpy (&pixDist0, &(lineDist[i-1]), sizeof(LineDistance));
		memcpy (&pixDist1, &(lineDist[i]), sizeof(LineDistance));

		// We consider the segment from pixDist0 to pixDist1.

		// Compute the error of this segment relative to the
		// "true" line segment as specified by clippedSegPixelCoords


		// First, generate overshoot error.  This routine also may
		// modify pixDist0, or pixDist1 to clip away the part
		// of the pixel segment that is beyond the limits of the
		// of the true segment.
#if 1
		calcOvershoot (segLen, &pixDist0, &pixDist1, 
				&overshootError, &signedOvershootError);
		*signedLineError += signedOvershootError;
		*lineError +=  overshootError;
#endif


		// If there is any of the pixel segment left, after
		// clipping out the overshoot portion, we find the area
		// between pixel segment and the true segment.

		if (pixDist0.alongLine < pixDist1.alongLine)
		    {
		    if ((pixDist1.toLine*pixDist0.toLine) >= 0)
			{
			// The two points are on the same side of the line
			// so we will treat as a trapezoid.

			signedArea = 
			  (pixDist1.alongLine - pixDist0.alongLine ) *
			  		(pixDist1.toLine + 
                           		 pixDist0.toLine ) / 2.0;
			unsignedArea = fabs (signedArea);
			}
		    else
			{
			// the points are on opposite sides of line
			// so we need to sum two triangles

			signedTri0 = pixDist0.toLine * 
			  (pixDist1.alongLine - pixDist0.alongLine ) *
				(fabs (pixDist0.toLine) /
				 (fabs (pixDist0.toLine) +
				 fabs (pixDist1.toLine)));

			signedTri1 = pixDist1.toLine * 
			  (pixDist1.alongLine - pixDist0.alongLine ) *
				(fabs (pixDist1.toLine) /
				 (fabs (pixDist0.toLine) +
				  fabs (pixDist1.toLine)));


			signedArea = signedTri0 + signedTri1;
			unsignedArea = fabs(signedTri0) + fabs(signedTri1);
			}


		    *signedLineError += signedArea;
		    *lineError += unsignedArea;
		    }
		}


#if 1
	// Now we look for undershoot and add its error into the total error.
	calcUndershoot (segLen, &(lineDist[0]), 
		&(lineDist[numRenderedPixels-1]), 
		&undershootError, &signedUndershootError);
	*lineError += (undershootError);
	*signedLineError += signedUndershootError;
#endif

	return 0;
	} // end of calculateErrors


static int 
getSegmentRenderErrors (
  View *view, 
  Segment seg, 
  double *signedPtError,
  double *ptError, 
  double *signedLineError,
  double *lineError, 
  int *numRenderedPixels,
  double *visibleScreenLineLen, 
  double *visibleWorldLineLen)
	{



	/*

	Here's what we need to do

		clip segment to viewing frustum

		determine length of clipped segment in world coords 
			(visibleWorldLineLen)

		project segment to pixel coords 

		determine length of projected clipped seg in pixel coords
			(visibleScreenLineLen)

		render segment using OpenGL

		get set of rendered pixels

		determine distance of each pixel center to projected
			clipped segment

		sum these to get ptError

		regard sequence of centers of rendered pixels as a 
			polyline and integrate distance from this
			polyline to clipped projected segment over
			length of clipped projected segment.  
			This is lineError

	*/
			







	int renderedSegVisible, calculatedSegVisible;
	int i, rtn;
	double len, a;
	static int renderedPixelCoords[MAX_NUM_RENDERED_PIXELS][2];
	Segment clippedSeg, clippedSegPixelCoords;


#define TRUE_VISIBLE 		0
#define FALSE_VISIBLE 		1
#define TRUE_INVISIBLE 		2
#define FALSE_INVISIBLE 	3



#if 0
if (Debug)
{
printf ("Starting seg  %f %f %f      %f %f %f\n",
seg[0][0],
seg[0][1],
seg[0][2],
seg[1][0],
seg[1][1],
seg[1][2]);
}
#endif

	calculatedSegVisible = 
	  ! clipToViewingFrustum (
		view, 
		seg, 
		clippedSeg , clippedSegPixelCoords);


#if 0
if (Debug)
{
printf (" clipped seg  %f %f %f      %f %f %f\n",
clippedSeg[0][0],
clippedSeg[0][1],
clippedSeg[0][2],
clippedSeg[1][0],
clippedSeg[1][1],
clippedSeg[1][2]);

printf (" clipped seg  pixel coords %f %f %f      %f %f %f\n",
clippedSegPixelCoords[0][0],
clippedSegPixelCoords[0][1],
clippedSegPixelCoords[0][2],
clippedSegPixelCoords[1][0],
clippedSegPixelCoords[1][1],
clippedSegPixelCoords[1][2]);
}
#endif


	renderedSegVisible = 
	  ! getRenderedPixels (
		view, 
		seg,
		numRenderedPixels, 
		renderedPixelCoords );

	renderedSegVisible = renderedSegVisible && (*numRenderedPixels > 0);

	if (calculatedSegVisible && renderedSegVisible)
		{
		calculateErrors 
			(
			clippedSegPixelCoords, 
			*numRenderedPixels, 
			renderedPixelCoords, 
			signedPtError,
			ptError, 
			signedLineError,
			lineError
			);

#if 0
moved below
		// calculate visibleScreenLineLen
		len = 0;
		for (i = 0; i < 2; i++)
			{
			a = clippedSegPixelCoords[0][i] - 
				clippedSegPixelCoords[1][i];
			len += a*a;
			}
		*visibleScreenLineLen = sqrt (len);


		// calculate visibleWorldLineLen
		len = 0;
		for (i = 0; i < 3; i++)
			{
			a = clippedSeg[0][i] - clippedSeg[1][i];
			len += a*a;
			}

		*visibleWorldLineLen = sqrt (len);
#endif

		rtn = TRUE_VISIBLE;
		}
	else if ((!calculatedSegVisible) && renderedSegVisible)
		{
		rtn = FALSE_VISIBLE;
		}
	else if ((!calculatedSegVisible) && (!renderedSegVisible))
		{
		rtn = TRUE_INVISIBLE;
		}
	else if ((calculatedSegVisible) && (!renderedSegVisible))
		{
		rtn = FALSE_INVISIBLE;
		}



	if (calculatedSegVisible)
		{

		// These derived values are valid whenever we
		// have a segment that intersects the view frustum.

		// calculate visibleScreenLineLen
		len = 0;
		for (i = 0; i < 2; i++)
			{
			a = clippedSegPixelCoords[0][i] - 
				clippedSegPixelCoords[1][i];
			len += a*a;
			}
		*visibleScreenLineLen = sqrt (len);


		// calculate visibleWorldLineLen
		len = 0;
		for (i = 0; i < 3; i++)
			{
			a = clippedSeg[0][i] - clippedSeg[1][i];
			len += a*a;
			}

		*visibleWorldLineLen = sqrt (len);
		}


	return rtn;
	}	// end of getPointRenderError
	



void init (void) 
{
   glClearColor (0.0, 0.0, 0.0, 0.0);
   glShadeModel (GL_FLAT);

   initRenderingTestCases ();
}


static FILE *ViewFP = NULL;
static FILE *ErrorFP = NULL;

void display(void)
	{

	View view;
	double ptError, lineError, signedPtError, signedLineError,
			visibleScreenLineLen, visibleWorldLineLen;
	Segment seg;
	int rtn;
	int numRenderedPixels;
	static int lastViewId = -1;
	

	// ViewFP = stdout;
	// ErrorFP = stdout;



	// Clear out the entire window.
	glViewport (0, 0, CurrentWindowSize[0], CurrentWindowSize[1] );
sleep (2);
	glClear (GL_COLOR_BUFFER_BIT);
sleep (2);
	glFinish ();



	while (getNextRenderingTestCase (&view, seg) == 0)
		{


		signedPtError =
		ptError =
		signedLineError =
		lineError =
		numRenderedPixels =
		visibleScreenLineLen = 
		visibleWorldLineLen = 0;

		rtn = getSegmentRenderErrors ( 
			&view, 
			seg, 
			&signedPtError,
			&ptError, 
			&signedLineError,
			&lineError,
			&numRenderedPixels, 
			&visibleScreenLineLen, 
			&visibleWorldLineLen );

#if 0
		printf (" 00%d "
			"%3d %3d "
			"%f "
			"%f %f %f %f "
			"%lf %lf %lf "
			"%lf %lf %lf "
			"%lf %lf "
			"%lf %lf "
			"%d "
			"%lf %lf "
			"\n",
			rtn,
			view.windowSize[0], view.windowSize[1], 
			view.camera.fov,
			view.camera.orientation[0],
			view.camera.orientation[1],
			view.camera.orientation[2],
			view.camera.orientation[3],
			seg[0][0],
			seg[0][1],
			seg[0][2],
			seg[1][0],
			seg[1][1],
			seg[1][2],
			signedPtError, ptError, 
			signedLineError, lineError,
			numRenderedPixels,
			visibleScreenLineLen,
			visibleWorldLineLen );



#else
		if (view.idNum != lastViewId)
			{
			printView (ViewFP, &view);
			lastViewId = view.idNum;
			}

		fprintf (ErrorFP,
			"%d  "
			"%d  "
			"%lf %lf %lf "
			"%lf %lf %lf "
			"%lf %lf "
			"%lf %lf "
			"%d "
			"%lf %lf "
			"\n",
			view.idNum,
			rtn,
			seg[0][0],
			seg[0][1],
			seg[0][2],
			seg[1][0],
			seg[1][1],
			seg[1][2],
			signedPtError, ptError, 
			signedLineError, lineError,
			numRenderedPixels,
			visibleScreenLineLen,
			visibleWorldLineLen );


#endif

#if 0
if ((rtn == 0))
{
Debug = 1;
rtn = getSegmentRenderErrors ( 
	&view, 
	seg, 
	&signedPtError,
	&ptError, 
	&signedLineError,
	&lineError,
	&numRenderedPixels, 
	&visibleScreenLineLen, 
	&visibleWorldLineLen );
Debug = 0;
}
#endif



#if 0
if (CurrentCase == 2)
{
sleep (30);
}
#endif
		}

	exit (0);

	}  // end of display


void reshape (int w, int h)
{



   // printf ("reshape %d %d\n", w, h);
   CurrentWindowSize[0] = w;
   CurrentWindowSize[1] = h;

}

void keyboard(unsigned char key, int x, int y)
{
   switch (key) {
      case 27:
         exit(0);
         break;
   }
}


static void
usage ()
	{
	fprintf (stderr, "\n");
	fprintf (stderr, 
	  "Usage: lineRenderingErrors [glut args] numSeg  viewFN  errorFN\n");
	fprintf (stderr,
	 "   [glut args] See GLUT documentation.\n");
	fprintf (stderr,
	 "   numSeg   number of line segments drawn for each view\n");
	fprintf (stderr,
	 "   viewFN   file to which a description of each view is written.\n");
	fprintf (stderr,
	 "   errorFN  file to which error data is written for each "
	 "segment and each view.\n");
	fprintf (stderr, "\n");
	return;
	}

//
// command line:    progName  numSeg  viewFN  errorFN
//

int main(int argc, char** argv)
{
   glutInit(&argc, argv);


   if (argc != 4)
	{
	fprintf (stderr, "Bad command line arguments.\n");
	usage ();
	exit (-1);
	}

   if (sscanf (argv[1], "%d", &NumTestSegments) != 1)
	{
	fprintf (stderr, "Bad argument: < %s >.\n", argv[1]);
	usage ();
	exit (-1);
	}

   if (NumTestSegments < 1)
	{
	fprintf (stderr, "Bad argument: < %s >.\n", argv[1]);
	usage ();
	exit (-1);
	}

   if ( (ViewFP = fopen (argv[2], "w")) == NULL)
	{
	fprintf (stderr, "Bad argument: < %s >.\n", argv[2]);
	usage ();
	exit (-1);
	}

   if ( (ErrorFP = fopen (argv[3], "w")) == NULL)
	{
	fprintf (stderr, "Bad argument: < %s >.\n", argv[3]);
	usage ();
	exit (-1);
	}



   // It seems to be faster (on acadie) to do it all in the front buffer
   glutInitDisplayMode (GLUT_SINGLE | GLUT_RGB);


   glutCreateWindow (argv[0]);

   init ();

   glutDisplayFunc(display); 
   glutReshapeFunc(reshape);
   glutKeyboardFunc(keyboard);

#if 0
   glutFullScreen ();
#else
   glutReshapeWindow (907, 907);
#endif


   glutMainLoop();
   return 0;
}
