
#define mantissa_bits 24
#define PRINT_VALS 0

/* Comments 1/89 (J. Devaney):
   This random number generator is a 'C' version of the FORTRAN code 
   provided with and described in the book "Numerical Methods and 
   Software" by D. Kahaner, C. Moler, and S. Nash, Prentice Hall, 1988.  
   It returns a uniform random number on [0,1).
  */

static float u[17] =
     {
      0.8668672834288,  0.3697986366357,  0.8008968294805,
      0.4173889774680,  0.8254561579836,  0.9640965269077,
      0.4508667414265,  0.6451309529668,  0.1645456024730,
      0.2787901807898,  0.06761531340295, 0.9663226330820,
      0.01963343943798, 0.02947398211399, 0.1636231515294,
      0.3976343250467,  0.2631008574685,
     };

float uni()

#define denom  16777216.0
#define csave  362436.0/denom
#define cd     7654321.0/denom
#define cm     16777213.0/denom


{

   static float c = csave;   
   static int   i = 16;
   static int   j = 4;
   float  rand;
   void circuit_breaker();



 /* Basic generator is Fibonacci
  */

   rand = u[i] - u[j];
   if (rand < 0.0) rand = rand + 1.0;
   u[i] = rand;
   i = i - 1;
   if (i == -1) i = 16;
   j = j - 1;
   if (j == -1) j = 16;
   


/* Second generator is congruential
 */

  c = c - cd;   
  if (c < 0.0) c = c + cm;



/* Combination generator
 */

  rand = rand - c;        
  if (rand < 0.0) rand = rand + 1.0;

  if ((rand < 0.0) || (rand >= 1.0))
     circuit_breaker("uni: Random number out of range.");
                                
  return(rand);

}


float ustart(iseed)
int iseed;

{
                                   
 float  s,t;
 int    ii,jj,i1,j1,k1,l1,m1;

#if PRINT_VALS == 1
printf(" %11.8f %11.8f %11.8f \n",u[0],u[1],u[2]);
printf(" %11.8f %11.8f %11.8f \n",u[3],u[4],u[5]);
printf(" %11.8f %11.8f %11.8f \n",u[6],u[7],u[8]);
printf(" %11.8f %11.8f %11.8f \n",u[9],u[10],u[11]);
printf(" %11.8f %11.8f %11.8f \n",u[12],u[13],u[14]);
printf(" %11.8f %11.8f        \n",u[15],u[16]);
#endif


 /* set up ...
   convert iseed to four smallish positive integers
 */

  i1 = abs(iseed) % 177;
  i1++;

  j1 = abs(iseed) % 167;
  j1++;

  k1 = abs(iseed) % 157;
  k1++;

  l1 = abs(iseed) % 147;
  l1++;

/* Generate random bit pattern in array based on given seed
 */

for ( ii = 0; ii < 17; ii++)
  { s = 0.0;
    t = 0.5;
    for ( jj = 1; jj <= mantissa_bits; jj++)
      { m1 = (((i1*j1) % 179)*k1) % 179;
        i1 = j1;
        j1 = k1;
        k1 = m1;
        l1 = (53*l1 + 1) % 169;
        if ( ((l1*m1) % 64) >= 32 ) s += t;
        t = 0.5*t;
      };
    u[ii] = s;
   };
 

#if PRINT_VALS == 1
printf(" %11.8f %11.8f %11.8f \n",u[0],u[1],u[2]);
printf(" %11.8f %11.8f %11.8f \n",u[3],u[4],u[5]);
printf(" %11.8f %11.8f %11.8f \n",u[6],u[7],u[8]);
printf(" %11.8f %11.8f %11.8f \n",u[9],u[10],u[11]);
printf(" %11.8f %11.8f %11.8f \n",u[12],u[13],u[14]);
printf(" %11.8f %11.8f        \n",u[15],u[16]);
#endif

s = iseed;
return(s);
}


#define ANSI 0

#if ANSI == 1
void circuit_breaker(char message[])  /* stop program */
#endif
#if ANSI == 0
void circuit_breaker(message)
char message[];
#endif
{
   fprintf(stderr,"%s\n",message);
   fprintf(stderr,"CRASH!!! BANG!!! BOOM!!! \n"); /* Dr. Strangelove */
   fprintf(stderr,"We'll meet again - don't know where, don't know when.\n");
   exit(1);
} /* end circuit_breaker */


